#include <assert.h>
#include <string.h>
#include "real.h"
#include DISPLAY

buf_t buffers[MAXBUFFERS];
buf_t *cb;
int bufw=0, bufh=0, scrh=0, bufn=0;

buf_cell_t ** allocbufdata()
{
	buf_cell_t ** d;
	d = (buf_cell_t **) calloc(CFG_BUFFER_SCROLLBACK + scrh, sizeof(buf_cell_t *));
	assert(d != NULL);
	return d;
}

/* display->buffer: set screen dimensions (visible buffer size) */
void buf_set_screen_dim(const int w, const int h)
{
	int olh = CFG_BUFFER_SCROLLBACK + scrh;
	int n, i, off, ol;
	buf_cell_t ** nd;

	scrh = h;

	off = olh - (CFG_BUFFER_SCROLLBACK+scrh);

	for(n=0; n<bufn; n++) {
		nd = allocbufdata();

		for(i = 0; i<CFG_BUFFER_SCROLLBACK+scrh; i++) {
			ol = i + off;
			if(ol >= 0) {
				ol = (i + off + buffers[n].loff) % olh;
				if(buffers[n].bufdata[ol]) {
					nd[i] = (buf_cell_t *) calloc(w, sizeof(buf_cell_t));
					assert(nd[i] != NULL);
					memcpy(nd[i], buffers[n].bufdata[ol], (w < bufw ? w : bufw) * sizeof(buf_cell_t));
					free(buffers[n].bufdata[ol]);
				}
			}
		}

		free(buffers[n].bufdata);
		buffers[n].bufdata = nd;
		buffers[n].loff = 0;
		buffers[n].bufstate.w = w;
		buffers[n].bufstate.h = scrh;
		buffers[n].bufstate.cursor_y -= off;
	}

	bufw = w;
	bufh = CFG_BUFFER_SCROLLBACK+scrh;

	for(n=0; n<bufn; n++)
		buf_offact(buffers+n);

	log_printf(2, "bufw=%d bufh=%d scrh=%d\n", bufw, bufh, scrh);

	if(cb) buf_update(cb, 0, 0, bufw-1, scrh-1);
}


/* scroll the buffer. bottom line will be filled with 0, default fg/bg/attr */
void buf_scroll(buf_t *b)
{
	buf_cursor_hide(b);

	if(b->bufdata[b->loff]) free(b->bufdata[b->loff]);
	b->bufdata[b->loff] = 0;
	b->loff = (b->loff + 1) % bufh;
	buf_offact(b);

	if(b == cb) display_scroll();
}

/* invalidate display */
void buf_update(buf_t *b, const int x1, const int y1, const int x2, const int y2)
{
	if(b == cb) display_update(x1, y1, x2, y2);
}

/* select active buffer */
void buf_select(buf_t *b)
{
	cb = b;
	display_set_buffer(cb);
}

/* switch buffer relatively */
void * buf_switch_rel(const int o)
{
	int idx = cb - buffers;
	idx += (o % bufn);
	if(idx < 0) idx += bufn;
	idx = idx % bufn;
	log_printf(2, "SELECT BUFFER: %d\n", idx);
	buf_select(buffers + idx);
	return cb;
}

/* display->buffer: creates new buffer, returns handle.
 * buffer will create emu and io */
void *buf_create()
{
	assert(bufn < MAXBUFFERS);

	buffers[bufn].bufstate.w = bufw;
	buffers[bufn].bufstate.h = scrh;
	buffers[bufn].bufstate.cursor_x = 0;
	buffers[bufn].bufstate.cursor_y = 0;
	buffers[bufn].bufstate.cursor_type = buf_cursor_block;
	buffers[bufn].loff = 0;
	buffers[bufn].bufdata = allocbufdata();
	buffers[bufn].bufstate.scrv = 0;
	buf_reset_color(buffers+bufn);
	buf_set_rw(buffers+bufn, 0, 0);

	buffers[bufn].bufemu = emu_create(buffers+bufn);
	buf_offact(buffers+bufn);

	bufn++;

	return buffers+bufn-1;
}

/* adjust buffer scroll */
void buf_adjust_scroll(buf_t *b, int amount)
{
	b->bufstate.scrv += amount;
	if(b->bufstate.scrv > CFG_BUFFER_SCROLLBACK) b->bufstate.scrv = CFG_BUFFER_SCROLLBACK;

	while(b->bufstate.scrv > 0 && !b->bufdata[(bufh + b->loff - scrh - b->bufstate.scrv) % bufh]) {
		b->bufstate.scrv --;
	}

	if(b->bufstate.scrv < 0) b->bufstate.scrv = 0;

	if(b == cb) display_update(0, 0, bufw-1, scrh-1);
}

/* death callback of buffer */
void buf_dead(buf_t *b)
{
}

/* erase part of buffer */
void buf_erase(buf_t *b, const int x1, const int y1, const int x2, const int y2)
{
	int x, y;
	assert(x1 <= x2 && y1 <= y2 && x2 < b->bufstate.w && y2 < b->bufstate.h);
	for(y=y1; y<=y2; y++) {
		buf_set_rw(b, x1, y);
		for(x=x1; x<=x2; x++)
			buf_write_char(b, 0);
	}

	buf_update(b, x1, y1, x2, y2);
}

/* shift part of buffer */
void buf_shift(buf_t *b, const int x1, const int y1, const int x2, const int y2, const int xd, const int yd)
{
	int fx, fy, tx, ty, x, y;
	int sx=0, sy=0;
	buf_cell_t c;

	if(xd < 0) {
		sx = 1;
		fx = x1; tx = x2;
	} else {
		sx = -1;
		fx = x2; tx = x1;
	}

	if(yd < 0) {
		sy = 1;
		fy = y1; ty = y2;
	} else {
		sy = -1;
		fy = y2; ty = y1;
	}

	ty += sy;
	tx += sx;

// TODO: speed up with memmove?
	for(y = fy; y != ty; y += sy) {
		for(x = fx; x != tx; x += sx) {
			buf_set_rw(b, x, y);
			c = buf_read_cell(b);
			buf_set_rw(b, x + xd, y + yd);
			*(b->rwptr) = c;
		}
	}

	if(b == cb) {
		display_cursor_hide();
		display_shift(x1, y1, x2, y2, xd, yd);
	}
}

