#include "timer.h"

typedef struct display_timer_s display_timer_t;

struct display_timer_s {
	int ttl;
	void (*callback)(display_timer_cmd_t cmd, void *cookie);
	void *cookie;
	display_timer_t *prev, *next;
};

static display_timer_t *timers;

void *display_timer_create(int ttl, void (*callback)(display_timer_cmd_t cmd, void *cookie), void *cookie)
{
	display_timer_t *t;

	t = malloc(sizeof(display_timer_t));

	t->ttl      = ttl;
	t->callback = callback;
	t->cookie   = cookie;
	t->prev     = NULL;
	t->next     = timers;

	timers = t;
	callback(timer_created, cookie);
	return t;
}

void display_timer_delete(void *timer) {
	display_timer_t *t = timer;

	/* Unlink from the double linked list */
	if (t->prev != NULL)
		t->prev->next = t->next;
	else
		timers = t->next;

	if (t->next != NULL)
		t->next->prev = t->prev;

	t->callback(timer_free, t->cookie);
	free(t);
}

void display_timer_run()
{
	display_timer_t *t;
	display_timer_t *tn;

	for(t = timers; t != NULL; t = tn) {
		tn = t->next;
		t->ttl--;
		if (t->ttl == 0) {
			t->callback(timer_run, t->cookie);
			display_timer_delete(t);
		}
	}
}


void display_timer_set_ttl(void *timer, int ttl) {
	display_timer_t *t = timer;
	t->ttl = ttl;
}
