#include <stdio.h> // DEBUG
#include <stdlib.h>

#include CONFIG
#include "dummy.h"
#include "global.h"

void *emu_create(void *buf)
{
	emu_dummy_t *emu = malloc(sizeof(emu_dummy_t));
	if (emu == NULL) {
		return NULL;
	}
	emu->es = DUMMY;
	emu->buf = buf;
	emu->bs = buf_get_state(buf);
	buf_set_cursor(emu->buf, 0, 0);
	emu->io = io_create(emu);
	return (void *)emu;
}

void emu_destroy(void *e)
{
	io_destroy(((emu_dummy_t *)e)->io);
	free(e);
}

// echo / write non control char
static inline void emu_write_char(emu_dummy_t *emu, emu_char_t c)
{
//	log_printf(2, "EMU_WRITE_CHAR('%c' %d), POS: %d %d\n", c, c, emu->bs->cursor_x, emu->bs->cursor_y);
	buf_rw_from_cursor(emu->buf);
	buf_write_char(emu->buf, c);
	if (buf_cursor_eor(emu->buf)) {
		if (buf_cursor_eoc(emu->buf)) {
			buf_scroll(emu->buf);
			buf_update(emu->buf, 0, emu->bs->h,  emu->bs->w - 1, emu->bs->h - 1);
			buf_set_cursor(emu->buf, 0, emu->bs->cursor_y);
		} else {
			buf_update(emu->buf, 0, emu->bs->cursor_y - 1,  emu->bs->w - 1, emu->bs->cursor_y);
			buf_set_cursor(emu->buf, 0, emu->bs->cursor_y + 1);
		}
	} else {
		buf_cursor_right(emu->buf);
		buf_update(emu->buf, emu->bs->cursor_x - 1, emu->bs->cursor_y,  emu->bs->cursor_x, emu->bs->cursor_y);
	}
}

static inline void emu_lf(emu_dummy_t *emu) {
	if (buf_cursor_eoc(emu->buf)) {
		buf_set_cursor(emu->buf, 0, emu->bs->cursor_y);
		buf_scroll(emu->buf);
	} else {
		buf_set_cursor(emu->buf, 0, emu->bs->cursor_y + 1);
	}
}

void emu_dsp_char(void *e, emu_char_t c) {
	emu_dummy_t *emu = e;
	io_got_user_input(emu->io, c);
}

void emu_io_char(void *e, emu_char_t c) {
	emu_dummy_t *emu = e;
	switch ((emu_control_char_t)c) {
		case LF :
			emu_lf(emu);
			break;
		case BS :
			if (emu->bs->cursor_x > 0)
				buf_cursor_left(emu->buf);
			buf_update(emu->buf, emu->bs->cursor_x, emu->bs->cursor_y,  emu->bs->cursor_x + 1, emu->bs->cursor_y);
			break;
		case CR :
			break;
		case BEL :
			display_bell();
			break;
		case HT :
			while((emu->bs->cursor_x % CFG_EMULATION_TAB) != 0)
				emu_write_char(emu, ' ');
			break;
		default:
			emu_write_char(emu, c);
	}
}

void emu_dsp_ctrl(void *e, ctrl_t ctrl)
{
	emu_dummy_t *emu = e;
	io_control(emu->io, ctrl);
}

void emu_setup_fd(int fd)
{

}
