#include <assert.h>
#include <stdio.h>
/*

terminal actions (uses buffer module's interface)

*/

static inline void emu_clear_screen(emu_t *e)
{
	emu_clear_lines(e, 0, e->bs->h - 1);
}

static inline void emu_clear_lines(emu_t *e, int top, int bottom)
{
	assert(top <= bottom && bottom < e->bs->h);
	buf_erase(e->buf, 0, top, e->bs->w - 1, bottom);
}

static inline void emu_clear_chars(emu_t *e, int start, int stop)
{
	assert(start <= stop && stop < e->bs->w);
	buf_erase(e->buf, start, e->bs->cursor_y, stop, e->bs->cursor_y);
}

// delete lines
static inline void emu_scrollup(emu_t *e, int top, int bottom, unsigned int n)
{
	if (bottom >= e->bs->h || top >= bottom || n < 1) {
		return;
	}
	if (n > bottom - top) {
		emu_clear_lines(e, top, bottom);
		buf_update(e->buf, 0, top, e->bs->w - 1, bottom);
		return;
	}
//	log_printf(2, "scrollup(%d, %d, %d)\n", top, bottom, n);
//	if(top == 0 && bottom == e->bs->h - 1 && n == 1) {
//		buf_scroll(e->buf);
//	} else {	
	buf_shift(e->buf, 0, top + n, e->bs->w - 1, bottom, 0, -n);
	emu_clear_lines(e, bottom - n + 1, bottom);
	buf_update(e->buf, 0, bottom, e->bs->w - 1, bottom);
//	}
	e->need_wrap = 0;
}

// insert lines
static inline void emu_scrolldown(emu_t *e, int top, int bottom, unsigned int n)
{
	if (bottom >= e->bs->h || top >= bottom || n < 1) {
		return;
	}
	if (n > bottom - top) {
		emu_clear_lines(e, top, bottom);
		buf_update(e->buf, 0, top, e->bs->w - 1, bottom);
		return;
	}
//	log_printf(2, "scrolldown(%d, %d, %d)\n", top, bottom, n);
	buf_shift(e->buf, 0, top, e->bs->w - 1, bottom - n, 0, n);
	emu_clear_lines(e, top, top + n - 1);
	buf_update(e->buf, 0, top, e->bs->w - 1, top);
	e->need_wrap = 0;
}

static inline void emu_insert_chars(emu_t *e, unsigned int n)
{
	if (n >= e->bs->w - e->bs->cursor_x)
		// TODO: clear?
		return;
	if (n < 1)
		return;
//	log_printf(2, "xy(%d,%d), insert(%d)\n", e->bs->cursor_x, e->bs->cursor_y, n);
	// TODO: buf shift in row
	buf_shift(e->buf, e->bs->cursor_x, e->bs->cursor_y, e->bs->w - n - 1, e->bs->cursor_y, n, 0);
	emu_clear_chars(e, e->bs->cursor_x, e->bs->cursor_x + n - 1);
	buf_update(e->buf, e->bs->cursor_x, e->bs->cursor_y, e->bs->cursor_x + n - 1, e->bs->cursor_y);
//	buf_update(e->buf, e->bs->cursor_x, e->bs->cursor_y, e->bs->w - 1, e->bs->cursor_y);
	e->need_wrap = 0;
}

static inline void emu_delete_chars(emu_t *e, unsigned int n)
{
	if (n >= e->bs->w - e->bs->cursor_x)
		// TODO: clear?
		return;
	if (n < 1)
		return;
//	log_printf(2, "xy(%d,%d), delete(%d)\n", e->bs->cursor_x, e->bs->cursor_y, n);
	buf_shift(e->buf, e->bs->cursor_x + n, e->bs->cursor_y, e->bs->w - 1, e->bs->cursor_y, - n, 0);
	emu_clear_chars(e, e->bs->w - n, e->bs->w - 1);
	buf_update(e->buf, e->bs->w - n, e->bs->cursor_y, e->bs->w - 1, e->bs->cursor_y);
	//need to show cursor:
	buf_update(e->buf, e->bs->cursor_x, e->bs->cursor_y, e->bs->cursor_x, e->bs->cursor_y);
	//buf_update(e->buf, e->bs->cursor_x, e->bs->cursor_y, e->bs->w - 1, e->bs->cursor_y);
	e->need_wrap = 0;
}

/*
 * gotoxy() must verify all boundaries, because the arguments
 * might also be negative. If the given position is out of
 * bounds, the cursor is placed at the nearest margin.
 */
static void emu_gotoxy(emu_t *e, int x, int y)
{
	if (x < 0)
		x = 0;
	else if (x >= e->bs->w)
		x = e->bs->w - 1;

	if (e->decom) {
		if (y < e->top)
			y = e->top;
		else if (y >= e->bottom)
			y = e->bottom - 1;
	} else {
		if (y < 0)
			y = 0;
		else if (y >= e->bs->h)
			y = e->bs->h - 1;
	}
	buf_set_cursor(e->buf, x, y);
	e->need_wrap = 0;
}

static void emu_gotoxay(emu_t *e, int x, int y)
{
	emu_gotoxy(e, x, e->decom ? (e->top + y) : y);
}

// TODO: scroll (Shift + PgUp/Dwn)
/*
static void emu_scrolldelta(emu_t *e, int lines)
{
	NI("");
}

static void emu_scrollback(emu_t *e, int lines)
{
	emu_linux_t *e = emu;
	if (!lines)
		lines = e->bs->h / 2;
	emu_scrolldelta(e, -lines);
}

static void emu_scrollfront(emu_t *e, int lines)
{
	emu_linux_t *e = emu;
	if (!lines)
		lines = e->bs->h / 2;
	emu_scrolldelta(e, lines);
}
*/

static inline void emu_lf(emu_t *e)
{
	if (e->bs->cursor_y == e->bottom) {
		emu_scrollup(e, e->top, e->bottom, 1);
	} else {
		buf_cursor_down(e->buf);
		/* TODO: better method for 'show cursor' */
		buf_update(e->buf, 0, e->bs->cursor_y, 0, e->bs->cursor_y);
	}
	e->need_wrap = 0;
}

static inline void emu_ri(emu_t *e)
{
	if (e->bs->cursor_y == e->top) {
		emu_scrolldown(e, e->top, e->bottom, 1);
	} else {
		buf_cursor_up(e->buf);
	}
	e->need_wrap = 0;
}

static inline void emu_cr(emu_t *e)
{
	/* TODO: buf move cursor in row */
	buf_set_cursor(e->buf, 0, e->bs->cursor_y);
	/* TODO: better method for 'show cursor' */
	buf_update(e->buf, 0, e->bs->cursor_y, 0, e->bs->cursor_y);
	e->need_wrap = 0;
}

static inline void emu_bs(emu_t *e)
{
	if (e->bs->cursor_x > 0) {
		buf_cursor_left(e->buf);
		e->need_wrap = 0;
	}
}

static inline void emu_del(emu_t *e)
{
}

static inline void emu_cursor_report(emu_t *e)
{
	char s[40];
	sprintf(s, "\033[%d;%dR", e->bs->cursor_y + (e->decom ? e->top + 1 : 1), e->bs->cursor_x + 1);
	emu_respond_string(e, s);
}

static inline void emu_status_report(emu_t *e)
{
	emu_respond_string(e, "\033[0n");	/* Terminal ok */
}

#define VT100ID "\033[?1;2c"
#define VT102ID "\033[?6c"
static inline void emu_respond_id(emu_t *e)
{
	emu_respond_string(e, VT102ID);
}

#if 0
static void mouse_report(emu_t *e, int butt, int mrx, int mry)
{
	char s[8];
	sprintf(s, "\033[M%c%c%c", (char)(' ' + butt), (char)('!' + mrx), (char)('!' + mry));
	respond_string(e, s);
}

/* invoked via ioctl(TIOCLINUX) and through set_selection */
static int mouse_reporting(void)
{
	return vc_cons[fg_console].d->vc_report_mouse;
}
#endif

static void emu_save_cur(emu_t *e)
{
	D("");
	e->saved_x = e->bs->cursor_x;
	e->saved_y = e->bs->cursor_y;
	e->saved_color = e->bs->color;
/*
	e->vc_s_charset	= e->vc_charset;
	e->vc_saved_G0		= e->vc_G0_charset;
	e->vc_saved_G1		= e->vc_G1_charset;
*/
}

static void emu_restore_cur(emu_t *e)
{
	D("");
	emu_gotoxy(e, e->saved_x, e->saved_y);
	buf_set_color(e->buf, e->saved_color);
/*
	e->vc_charset		= e->vc_s_charset;
	e->vc_G0_charset	= e->vc_saved_G0;
	e->vc_G1_charset	= e->vc_saved_G1;
	e->vc_translate	= set_translate(e->vc_charset ? e->vc_G1_charset : e->vc_G0_charset, vc);
*/
	e->need_wrap = 0;
}

