#define _XOPEN_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <assert.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <errno.h>
#define HAVE_IO
#include CONFIG
#include "pty.h"
#include DISPLAY
#include EMULATION
#include "global.h"
#include <string.h>
#include <asm/ioctls.h>
#include <pty.h>
#define IO_HAVE_RESIZE
#define IO_HAVE_SIGNALS
#include "../io_common.c"

/* Includes are broken (compared to the manuals) */
extern int posix_openpt(int flags);
extern int killpg(int group, int signal);
extern char *strdup(char *str);
extern int setpgrp(void);

/* Remember shell settings */
static char *shell = NULL;
static char **shell_args = NULL;

void io_set_shell(char *shell_, char **shell_args_)
{
	shell = strdup(shell_);
	if (shell_args_ == NULL) {
		shell_args = malloc(sizeof(char *) * 2);
		shell_args[0] = shell;
		shell_args[1] = NULL;
	}
	else
		shell_args = shell_args_;
}

extern void emu_setup_fd(int fd);

io_instance_t *io_create(void *emu)
{
	io_instance_t *i;
	pid_t pid;
	pid_t pgid;
	int n, tty;

	i = malloc(sizeof(io_instance_t));

	i->master = posix_openpt(O_RDWR);

	if (grantpt(i->master) != 0) {
		log_printf(1, "Couldn't grant permissions on pty (%d)\n", i->master);
		abort();
	}

	if (unlockpt(i->master) != 0) {
		log_printf(1, "Unable to unlock the master side of pty\n");
	}

	log_printf(2, "Opening: %s\n", ptsname(i->master));
	i->slave=open(ptsname(i->master), O_RDWR);
	assert(i->slave >= 0);

	signal(SIGCHLD, io_sigchld);

	fcntl(i->master, F_SETFL, O_NONBLOCK);
	emu_setup_fd(i->master);
	pid = fork();
	if (pid == 0) {
	  for(n = 0; n < 100 ; n++)
  	  if(n != i->slave)
    	  close(n);

		pgid = setsid();
		if (pgid < 0)
			log_printf(1, "start session failed");

		/* Redirect stdio */
		dup2(i->slave, 0);
		dup2(i->slave, 1);
		dup2(i->slave, 2);

		if(ioctl(i->slave, TIOCSCTTY, NULL) < 0) {
			tty = open("/dev/tty", O_RDWR);
			if (tty >= 0) {
				ioctl(tty, TIOCNOTTY, NULL);
				close(tty);
			}
    }

	  setpgrp();
		putenv("TERM=" TERM_NAME);
		if (shell == NULL)
			io_set_shell(CFG_IO_DEFAULT_SHELL, NULL);
		execv(shell, shell_args);
	}

	close(i->slave);

	assert(pid > 0);

	signal(SIGINT, SIG_IGN);
	signal(SIGTSTP, SIG_IGN);

	i->pid = pid;
	i->emu = emu;


	display_watch(i->master, 1, io_got_output, i);
	return i;
}

void io_destroy(io_instance_t *i)
{
	close(i->master);
	close(i->slave);
	free(i);
}

void io_control(io_instance_t *i, ctrl_t ctrl) {
	switch(ctrl) {
		case ctrl_hup:
			break;
		case ctrl_term:
			killpg(0, SIGTERM);
			break;
		case ctrl_resize:
			io_resize(i);
			break;
	}
}
